from general.exchange.type_enum import BINANCEUM, OKX
from general.exchange.binance_um import BinanceUMHandler
from general.exchange.okx import OKXHandler


class ExchangeFactory:
    @staticmethod
    def get_handler(exchange: str, *api_args, **kwargs):
        if not api_args:
            api_args = ['', '', '']  # walk around undetermined api_args
        if exchange == BINANCEUM:
            return BinanceUMHandler(api_args[0], api_args[1], kwargs.get('logger', None))
        elif exchange == OKX:
            return OKXHandler(api_args[0], api_args[1], api_args[2], kwargs.get('logger', None))
        else:
            raise ValueError(f"{exchange} not implemented yet")



from general.exchange.schema import (
    TradeRecorder, Order, Orderbook
)

def to_side(side: str):  # either 'buy' or 'sell'
    return side.upper()


def to_symbol(base: str, quote: str):
    return f'{base}{quote}'


def to_trade_recorder(record) -> TradeRecorder:
    return TradeRecorder(
        exchange='BinanceUM',
        timestamp=record['time'],
        base=record['symbol'][:-4],
        quote=record['symbol'][-4:],
        side=record['side'],
        price=float(record['price']),
        qty=float(record['qty'])
    )


def to_order(order) -> Order:
    return Order(
        exchange='BinanceUM',
        order_id=order['orderId'],
        status=order['status'],
        updated_time=order['updateTime'],
        base=order['symbol'][:-4],
        quote=order['symbol'][-4:],
        qty=0 if order['origQty'] == '' else float(order['origQty']),
        side=order['side'],
        filled=0 if order['executedQty'] == '' else float(order['executedQty']),
        price=0 if order['price'] == '' else float(order['price'])
    )


def to_orderbook(orderbook) -> Orderbook:
    return Orderbook(
        exchange='BinanceUM',
        timestamp=orderbook['T'],
        ask=[(float(price), float(qty)) for price, qty in orderbook['asks']],
        bid=[(float(price), float(qty)) for price, qty in orderbook['bids']]
    )



from dataclasses import dataclass
from typing import List, Tuple


@dataclass
class Orderbook:
    exchange: str
    timestamp: int
    ask: List[Tuple[float, float]]
    bid: List[Tuple[float, float]]




import redis


class RedisPubSubHandler:
    def __init__(self, host, port, db, logger):
        self.logger = logger
        self.redis_conn = redis.StrictRedis(host=host, port=port, db=db)
        self.pubsub = self.redis_conn.pubsub()

    def publish(self, channel, message):
        self.redis_conn.publish(channel, message)

    def subscribe(self, channels):
        if isinstance(channels, str):
            channels = [channels]
        for channel in channels:
            self.pubsub.subscribe(channel)

    def listen(self, callback=None):
        for message in self.pubsub.listen():
            if message['type'] == 'message':
                channel = message['channel']
                data = message['data']
                self.logger.debug(f"Received message '{data}' on channel '{channel}'")
                if callback:
                    callback(data)



import traceback
from okx.Trade import TradeAPI
from okx.MarketData import MarketAPI
from general.exchange.type_enum import OKX
from general.exchange.exchange_handler import ExchangeHandler
from general.exchange.type_enum import (
    GTC
)
from .utils import to_symbol, to_side, to_order, to_orderbook, to_trade_recorder


def exception_handler(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            args[0].logger.error(f"Missing key in message: {e}\n{traceback.format_exc()}")
    return wrapper


class OKXHandler(ExchangeHandler):
    NAME = OKX

    def __init__(self, api_key='', api_secret='', passphrase='', logger=None):
        super().__init__(logger)
        self.api_key = api_key
        self.api_secret = api_secret
        self.api_passphrase = passphrase
        self.okx_trade_demo = TradeAPI('30fb4bfc-d3ed-43ca-9656-ced5ea3e7a41',
                                       '9BFD52327F88CD251CE1FCEC0BFF495A',
                                       '2rYHFKwhxvtBKBZ$$', False, '1', debug=False)
        self.okx_market = MarketAPI(api_key, api_secret, passphrase, False, '0',
                                    domain='https://aws.okx.com', debug=False)
        self.okx_trade = TradeAPI(api_key, api_secret, passphrase, False, '0',
                                  domain='https://aws.okx.com', debug=False)

    @exception_handler
    def create_market_order(self, base, quote, side, qty, dry_run=False):
        if dry_run:
            res = self.okx_trade_demo.place_order(to_symbol(base, quote), 'cross', to_side(side),
                                                  'market', qty, posSide='long')  # not require in net mode
        else:
            res = self.okx_trade.place_order(to_symbol(base, quote), 'cross', to_side(side),
                                             'market', qty)
        return res

    @exception_handler
    def create_limit_order(self, base, quote, side, qty, price, time_in_force=GTC, dry_run=False):
        if dry_run:
            res = self.okx_trade_demo.place_order(to_symbol(base, quote), 'cross', to_side(side),
                                                  'limit', qty, px=price, posSide='long')
        else:
            res = self.okx_trade.place_order(to_symbol(base, quote), 'cross', to_side(side), 'limit',
                                             qty, px=price)
        return res

    @exception_handler
    def cancel_all_order(self, base, quote):
        orders = self.get_open_order(base, quote)
        inst = to_symbol(base, quote)
        order_id_list = [{"instId": inst, "ordId": order.order_id} for order in orders]
        ret = self.okx_trade.cancel_multiple_orders(order_id_list)
        return ret

    @exception_handler
    def get_open_order(self, base, quote):
        ret = self.okx_trade.get_order_list(instType='SWAP')
        res = filter(lambda order: order['instId'] == to_symbol(base, quote), ret['data'])
        return list(map(to_order, res))

    @exception_handler
    def get_orderbook(self, base, quote, limit=10):
        ret = self.okx_market.get_orderbook(to_symbol(base, quote), str(limit))
        return to_orderbook(ret['data'][0])

    @exception_handler
    def get_account_trades(self, base, quote, limit=100):
        ret = self.okx_market.get_trades(to_symbol(base, quote), limit=limit)
        res = list(map(to_trade_recorder, ret['data']))
        return res

    @staticmethod
    def to_market_price(data):
        data = data['data'][0]
        return {
            '_id': data['seqId'],
            'ex': OKXHandler.NAME,
            'b': data['bids'][0][0],    
            'a': data['asks'][0][0],
            't': int(data['ts'])
        }



import threading
import queue


class Worker(threading.Thread):
    def __init__(self, queue):
        super().__init__()
        self.queue = queue
        self.daemon = True  # Set as a daemon so it will close when the main thread closes

    def run(self):
        while True:
            try:
                # Get a job from the queue
                func, args = self.queue.get(timeout=3)  # Wait for 3 seconds to get a job
                if func is None:
                    # If the job is None, break the loop and end the worker
                    break
                # Execute the job with provided arguments
                func(*args)
            except queue.Empty:
                continue
            finally:
                # Mark the task as done in the queue
                self.queue.task_done()


class Parallelism:
    def __init__(self, n_workers):
        self.queue = queue.Queue()
        self.workers = []

        for _ in range(n_workers):
            worker = Worker(self.queue)
            self.workers.append(worker)
            worker.start()

    def add_job(self, func, args=()):
        self.queue.put((func, args))

    def wait_completion(self):
        # Wait until the queue is empty and all tasks are done
        self.queue.join()

        # Stop all worker threads after all tasks are done
        for _ in self.workers:
            self.queue.put((None, None))
        for worker in self.workers:
            worker.join()


import threading
import queue


class Worker(threading.Thread):
    def __init__(self, queue):
        super().__init__()
        self.queue = queue
        self.daemon = True  # Set as a daemon so it will close when the main thread closes

    def run(self):
        while True:
            # Get a job from the queue and block indefinitely until a job is available
            func, args = self.queue.get()

            # Execute the job with provided arguments
            func(*args)

            # Mark the task as done in the queue
            self.queue.task_done()


class ThreadPool:
    def __init__(self, n_workers):
        self.queue = queue.Queue()
        self.workers = []

        for _ in range(n_workers):
            worker = Worker(self.queue)
            self.workers.append(worker)
            worker.start()

    def add_job(self, func, args=()):
        self.queue.put((func, args))



import json
import redis


class RedisQueueHandler:
    def __init__(self, host, port, queue_name):
        self.redis_conn = redis.Redis(host=host, port=port, db=0)
        self.queue_name = queue_name

    def enqueue(self, data):
        self.redis_conn.rpush(self.queue_name, json.dumps(data))

    def dequeue(self):
        _, msg = self.redis_conn.blpop(self.queue_name)
        return json.loads(msg)

    def close(self):
        del self.redis_conn



import logging
from typing import Dict


class ExchangeHandler:
    def __init__(self, logger=None):
        if logger:
            self.logger = logger
        else:
            self.logger = logging.Logger(ExchangeHandler.__name__)

    def create_market_order(self, base, quote, side, qty, dry_run):
        pass

    def create_limit_order(self, base, quote, side, qty, price, time_in_force, dry_run):
        pass

    def cancel_all_order(self, base, quote):
        pass

    def get_orderbook(self, base, quote, limit):
        pass

    def get_open_order(self, base, quote):
        pass

    def get_account_trades(self, base, quote):
        pass

    @staticmethod
    def to_market_price(data: Dict) -> Dict:
        """
        Convert exchange orderbook data to schema
        {
            _id: int,
            ex: str,
            a: float
            b: float
            t: int
        }
        """
        return data



from influxdb_client import InfluxDBClient, Point
from influxdb_client.client.write_api import SYNCHRONOUS


class InfluxHandler:
    def __init__(self, url, token, batch_size=1000):
        self.client = InfluxDBClient(url=url, token=token)
        self.org = None
        self.bucket = None
        self.batch_size = batch_size
        self.points = []

    def set_org(self, org_name):
        self.org = org_name

    def set_bucket(self, bucket_name):
        self.bucket = bucket_name

    def add_point(self, measurement, tags, fields):
        if self.bucket is None:
            raise Exception("Bucket is not set. Use set_bucket() method first.")

        point = Point(measurement)
        for tag_key, tag_value in tags.items():
            point.tag(tag_key, tag_value)
        for field_key, field_value in fields.items():
            point.field(field_key, field_value)
        self.points.append(point)

        if len(self.points) >= self.batch_size:
            self.write_points()

    def write_points(self):
        if self.bucket is None or self.org is None:
            raise Exception("Bucket or Organization is not set. Use set_bucket() and set_org() methods first.")

        write_api = self.client.write_api(write_options=SYNCHRONOUS)
        write_api.write(bucket=self.bucket, org=self.org, record=self.points)
        self.points = []

    def query_data(self, flux_query):
        if self.org is None:
            raise Exception("Organization is not set. Use set_org() method first.")

        query_api = self.client.query_api()
        tables = query_api.query(org=self.org, query=flux_query)
        results = []
        for table in tables:
            for row in table.records:
                results.append(row.values)
        return results

    def close(self):
        if self.points:
            self.write_points()
        self.client.__del__()


from general.exchange.type_enum import BINANCEUM
from general.exchange.exchange_handler import ExchangeHandler
from binance.um_futures import UMFutures
from binance.error import ClientError
from general.exchange.type_enum import (
    MARKET, LIMIT,
    GTC
)
from .utils import (
    to_side, to_symbol, to_trade_recorder, to_order, to_orderbook
)


def exception_handler(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ClientError as error:
            # Assuming the first argument to the function is `self`
            args[0].logger.error(
                "Found error. status: {}, error code: {}, error message: {}".format(
                    error.status_code, error.error_code, error.error_message
                )
            )
    return wrapper


class BinanceUMHandler(ExchangeHandler):
    NAME = BINANCEUM

    def __init__(self, api_key='', api_secret='', logger=None):
        super().__init__(logger)
        self.um_futures = UMFutures(key=api_key, secret=api_secret)

    @exception_handler
    def create_market_order(self, base, quote, side, qty, dry_run=False):
        if dry_run:
            ret = self.um_futures.new_order_test(symbol=to_symbol(base, quote), side=to_side(side), type=MARKET,
                                                 quantity=qty)
        else:
            ret = self.um_futures.new_order(symbol=to_symbol(base, quote), side=to_side(side), type=MARKET,
                                            quantity=qty)
        return to_order(ret)

    @exception_handler
    def create_limit_order(self, base, quote, side, qty, price, time_in_force=GTC, dry_run=False):
        if dry_run:
            ret = self.um_futures.new_order_test(symbol=to_symbol(base, quote), side=to_side(side), type=LIMIT, quantity=qty, price=price,
                                                 timeInForce=time_in_force)
        else:
            ret = self.um_futures.new_order(symbol=to_symbol(base, quote), side=to_side(side), type=LIMIT, quantity=qty, price=price,
                                            timeInForce=time_in_force)
        return to_order(ret)

    @exception_handler
    def cancel_all_order(self, base, quote):
        ret = self.um_futures.cancel_open_orders(symbol=to_symbol(base, quote))
        return ret

    @exception_handler
    def get_orderbook(self, base, quote, limit=10):
        ret = self.um_futures.depth(symbol=to_symbol(base, quote), limit=limit)
        return to_orderbook(ret)

    @exception_handler
    def get_open_order(self, base, quote):
        ret = self.um_futures.get_all_orders(symbol=to_symbol(base, quote))
        ret = list(filter(lambda order: order["status"] == 'NEW', ret))
        return list(map(to_order, ret))

    @exception_handler
    def get_account_trades(self, base, quote):
        ret = self.um_futures.get_account_trades(symbol=to_symbol(base, quote))
        return list(map(to_trade_recorder, ret))

    @staticmethod
    def to_market_price(data):
        return {
            '_id': data['E'],
            'ex': BinanceUMHandler.NAME,
            'b': data['b'][0][0],
            'a': data['a'][0][0],
            't': data['T']
        }



from setuptools import setup, find_packages

setup(
    name='coin-infra',
    version='4.2.2',
    description='infra of coin projects',
    author='Yi Te',
    author_email='coastq22889@icloud.com',
    packages=find_packages(include=['general', 'general.*']),
    install_requires=[
        'pymongo',
        'redis',
        'requests',
        'websockets',
        'binance-futures-connector',
        'python-okx',
        'influxdb-client'
    ],
)





import time

from general.exchange.schema import (
    TradeRecorder, Order, Orderbook
)

def to_side(side: str):  # either 'buy' or 'sell'
    return side.lower()


def to_symbol(base: str, quote: str, instrument: str = 'SWAP'):
    return f'{base}-{quote}-{instrument}'


def to_trade_recorder(record) -> TradeRecorder:
    base, quote, inst = record['instId'].split('-')
    return TradeRecorder(
        exchange='OKX',
        timestamp=record['ts'],
        base=base,
        quote=quote,
        side=record['side'].upper(),
        price=float(record['px']),
        qty=float(record['sz'])
    )


def to_order(order) -> Order:
    base, quote, inst = order['instId'].split('-')
    return Order(
        exchange='OKX',
        order_id=order['ordId'],
        status=order['state'],
        updated_time=int(order['uTime']),
        base=base,
        quote=quote,
        qty=float(order['sz']),
        side=order['side'].upper(),
        filled=float(order['fillSz']),
        price=float(order['px'])
    )


def to_orderbook(orderbook) -> Orderbook:
    return Orderbook(
        exchange='OKX',
        timestamp=orderbook['ts'],
        ask=[(float(price), float(qty)) for price, qty, _, _ in orderbook['asks']],
        bid=[(float(price), float(qty)) for price, qty, _, _ in orderbook['bids']],
    )



from dataclasses import dataclass


@dataclass
class Order:
    exchange: str
    order_id: str
    status: str
    updated_time: int
    base: str
    quote: str
    qty: float
    side: str
    filled: float
    price: float



from pymongo import MongoClient


class MongoHandler:
    def __init__(self, host, port, username=None, password=None, auth_db="admin"):
        self.client = self._connect(host, port, username, password, auth_db)
        self.db = None
        self.collection = None

    def _connect(self, host, port, username, password, auth_db):
        if username and password:
            connection_uri = f"mongodb://{username}:{password}@{host}:{port}/{auth_db}"
        else:
            connection_uri = f"mongodb://{host}:{port}/"
        return MongoClient(connection_uri)

    def set_database(self, db_name):
        self.db = self.client[db_name]

    def set_collection(self, collection_name):
        if self.db is None:
            raise Exception("Database is not set. Use set_database() method first.")
        self.collection = self.db[collection_name]

    def insert(self, data):
        if self.collection is None:
            raise Exception("Collection is not set. Use set_collection() method first.")
        return self.collection.insert_one(data).inserted_id

    def insert_many(self, data_list):
        if self.collection is None:
            raise Exception("Collection is not set. Use set_collection() method first.")
        return self.collection.insert_many(data_list).inserted_ids

    def find(self, query=None):
        if self.collection is None:
            raise Exception("Collection is not set. Use set_collection() method first.")
        return list(self.collection.find(query))
